/* 
 * Copyright (C) 2002 Laird Breyer
 *  
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 * 
 * Author:   Laird Breyer <laird@lbreyer.com>
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <ctype.h>
#include <string.h>
#include <stdlib.h>


#include "dbacl.h"

extern options_t options;

extern Regex re[MAX_RE];
extern regex_count_t regex_count;

/* global variables */

MBOX_State mbox;
XML_State xml;

char *textbuf = NULL;
charbuf_len_t textbuf_len = 0;
char *aux_textbuf = NULL;
charbuf_len_t aux_textbuf_len = 0;

#if defined HAVE_MBRTOWC
wchar_t *wc_textbuf = NULL;
charbuf_len_t wc_textbuf_len = 0;
#endif

token_order_t ngram_order = 1; /* default */


/* similar to strstr, but case insensitive */
const char *mystrcasestr(const char *haystack, const char *needle) {
  const char *p, *q, *r;

  for(p = haystack; *p; p++) {
    q = needle; r = p;
    while( *q && *r && ((tolower(*q) - tolower(*r)) == 0) ) {
      q++; r++;
    }
    if( !*q ) {
      return p;
    }
  }
  return NULL;
}

#if defined HAVE_MBRTOWC
/* compiler doesn't seem to know this function is in the 
 * library, so we define our own - bug or just plain weird? */
int wcsncasecmp(const wchar_t *s1, const wchar_t *s2, size_t n) {
  register size_t i = 0;
  while( i < n ) {
    if( tolower(*s1) != tolower(*s2) ) {
      return towlower(*s1) - towlower(*s2);
    }
    s1++;
    s2++;
  }
  return 0;
}

/* similar to strstr, but case insensitive */
const wchar_t *mywcscasestr(const wchar_t *haystack, const wchar_t *needle) {
  const wchar_t *p, *q, *r;

  for(p = haystack; *p; p++) {
    q = needle; r = p;
    while( *q && ((towlower(*q) - towlower(*r)) == 0) ) {
      q++; r++;
    }
    if( !*q ) {
      return p;
    }
  }
  return NULL;
}

#endif


void init_file_handling() {
  /* preallocate primary text holding buffer */
  textbuf_len = BUFLEN;
  textbuf = malloc(textbuf_len);
}

void cleanup_file_handling() {
  /* free some global resources */
  free(textbuf);
}

void reset_mbox_line_filter() {
  mbox.state = msUNDEF;
  mbox.header.type = ctUNDEF;
  mbox.header.encoding = ceUNDEF;
  mbox.body.type = ctUNDEF;
  mbox.body.encoding = ceUNDEF;
  mbox.prev_line_empty = 1;
}

void reset_xml_character_filter() {
  xml.state = TEXT;
}

/***********************************************************
 * MULTIBYTE FILE HANDLING FUNCTIONS                       *
 * this is suitable for any locale whose character set     *
 * encoding doesn't include NUL bytes inside characters    *
 ***********************************************************/

bool_t good_char(char c) {
  if( options & (1<<OPTION_CHAR_ALNUM) ) {
    return isalnum((unsigned char)c);
  } else if( options & (1<<OPTION_CHAR_ALPHA) ) {
    return isalpha((unsigned char)c);
  } else if( options & (1<<OPTION_CHAR_GRAPH) ) {
    return isgraph((unsigned char)c);
  }
  return 0;
}

/* returns true if the line should be processed further
   depends on global mbox state */
/* note: mail filtering is very complicated, since
   any message can be mangled and requoted by a sender
   this will only cope with clean mail boxes */
bool_t mbox_line_filter(char *line) {

  bool_t line_empty = 0;
  bool_t process_line = 0; /* by default we skip the line */

  line_empty = ((line[0] == '\n') ||
		((line[0] == '\r') && (line[1] == '\n'))) ? 1 : 0;

  /* below we decide if we want to process the line */
  switch(mbox.state) {
  case msUNDEF:
    if( mbox.prev_line_empty && !strncmp(line, "From ", 5) ) {
      mbox.state = msHEADER;
      mbox.header.type = mbox.body.type = ctUNDEF;
      mbox.header.encoding = mbox.body.encoding = ceUNDEF;
    }
    break;
  case msHEADER:
    if( line_empty ) {
      mbox.state = msBODY;
      /* if the header was marked specially, transfer to body */
      mbox.body = mbox.header;
      /* actually, the first body after the header is always text */
      mbox.header.type = ctTEXT;
      mbox.header.encoding = ceID;
    } else {
      if( !strncasecmp(line, "Content-Type:", 13) ) {
	if( mystrcasestr(line + 13, "text") || 
	    mystrcasestr(line + 13, "message") ||
	    mystrcasestr(line + 13, "multipart") ) {
	  mbox.header.type = ctTEXT;
	} else {
	  mbox.header.type = ctOTHER;
	}
      } else if( !strncasecmp(line, "Content-Transfer-Encoding:", 26) ) {
	if( mystrcasestr(line + 36, "base64") ) {
	  mbox.header.encoding = ceB64;
	} else if( mystrcasestr(line + 36, "quoted-printable") ) {
	  mbox.header.encoding = ceQP;
	} else if( mystrcasestr(line + 36, "binary") ) {
	  mbox.header.encoding = ceBIN;
	} else {
	  mbox.header.encoding = ceID;
	}
      }
    }
    break; 
  case msBODY:
    if( mbox.prev_line_empty ) {
      if( (line[0] == '-') && (line[1] == '-') && !isspace((unsigned char)line[2]) ) {
	mbox.state = msATTACH;
	mbox.body.type = ctUNDEF;
	mbox.body.encoding = ceUNDEF;
      } else if( !strncmp(line, "From ", 5) ) {
	mbox.state = msHEADER;
	mbox.header.type = mbox.body.type = ctUNDEF;
	mbox.header.encoding = mbox.body.encoding = ceUNDEF;
      }
    }
    break;
  case msATTACH:
    if( line_empty ) {
      mbox.state = msBODY;
    } else {
      if( !strncasecmp(line, "Content-Type:", 13) ) {
	if( mystrcasestr(line + 13, "text") ) {
	  mbox.body.type = ctTEXT;
	} else {
	  mbox.body.type = ctOTHER;
	}
      } else if( !strncasecmp(line, "Content-Transfer-Encoding:", 26) ) {
	if( mystrcasestr(line + 36, "base64") ) {
	  mbox.body.encoding = ceB64;
	} else if( mystrcasestr(line + 36, "quoted-printable") ) {
	  mbox.body.encoding = ceQP;
	} else if( mystrcasestr(line + 36, "binary") ) {
	  mbox.body.encoding = ceBIN;
	} else {
	  mbox.body.encoding = ceID;
	}
      } 
    }
    break;
  }

  mbox.prev_line_empty = line_empty; /* for next time */

  /* now decide what to do about this line */
  switch(mbox.state) {
  case msHEADER:
    if( !strncasecmp(line, "From:",5) ) {
      process_line = 1;
    } else if( !strncasecmp(line, "Subject:",8) ) {
      /* this is where we could remove classifier tags */
      process_line = 1;
    }
    break;
  case msBODY:
    {
      mbox.scratch.type = 
	(mbox.body.type != ctUNDEF) ? mbox.body.type : mbox.header.type;
      if( mbox.scratch.type != ctOTHER ) {
	mbox.scratch.encoding = 
	  (mbox.body.encoding != ceUNDEF) ? 
	  mbox.body.encoding : mbox.header.encoding;
	if( (mbox.scratch.encoding == ceID) || 
	    (mbox.scratch.encoding == ceQP) ) {
	  process_line = 1;
	}
      }
    }
    break;
  case msATTACH:
    /* MIME section headers are always processed, since they describe content */
    process_line = 1;
    break;
  default:
    /* nothing */
    break;
  }

  return process_line;
}

/* removes tags in the string - modifies in place */
/* the name of this function is a misnomer, since it doesn't
   parse xml properly. But we just want a simple kludge 
   for most html flavours */
void xml_character_filter(char *line) {

  char *q;

  q = line;
  while( *line ) {
    switch(xml.state) {
    case TEXT:
      /* does it look like <x where x is either alpha or punctuation? */
      if( line[0] == '<' ) {
	if( !strncmp(line + 1, "!--", 3) ) {
	  xml.state = COMMENT;
	  line += 3;
	} else if( !strncasecmp(line + 1, "script", 6) ) {
	  xml.state = SPECIAL;
	  line += 6;
	} else if( !strncasecmp(line + 1, "br>", 3) ) {
	  *q++ = ' ';
	  line += 3;
	} else if( isalpha((unsigned char)line[1]) || ispunct((unsigned char)line[1]) ) {
	  xml.state = TAG;
	  line += 1;
	} else {
	  *q++ = *line;
	}
      } else {
	*q++ = *line;
      }
      break;
    case TAG:
      if( line[1] == '>' ) {
	xml.state = TEXT;
      }
      break;
    case COMMENT:
      if( line[0] == '-' && !strncmp(line + 1, "->", 2) ) {
	xml.state = TEXT;
	line += 2;
      }
      break;
    case SPECIAL:
      if( line[0] == '<' ) {
	if( !strncasecmp(line + 1, "/script", 7) ) {
	  xml.state = TEXT;
	  line += 7;
	}
      }
      break;
    }
    line++;
  }
  *q = '\0'; /* mark the end of the clean text string */
}

#define MULTIBYTE_EPSILON 10 /* enough for a multibyte char and a null char */

/* reads a text file as input, converting each line
into a wide character representation and applies several
filters. */
void process_file(FILE *input, 
		  int (*line_filter)(char *),
		  void (*character_filter)(char *), 
		  void (*word_fun)(char *, token_order_t, regex_count_t), 
		  char *(*pre_line_fun)(char *),
		  void (*post_line_fun)(char *)) {
  char *s, *pptextbuf;
  regex_count_t i;
  charbuf_len_t k;
  int eflag;
  token_order_t z, j, n, order;
  charbuf_len_t l;

  char *p, *q, *qq;
  char tok[MAX_TOKEN_LEN+2];

  regmatch_t pmatch[MAX_SUBMATCH];

  char ntok[MAX_TOKEN_LEN+2];
  char *nq;
  token_order_t nhow_many;

#if defined HAVE_LIBBOOST_REGEX
  charbuf_len_t wclen;
  wchar_t *wp;
  mbstate_t tok_shiftstate, input_shiftstate;
  charbuf_len_t tok_len;

  memset(&input_shiftstate, 0, sizeof(mbstate_t));
#endif


  /* initialize the norex state */
  ntok[0] = DIAMOND;
  ntok[1] = '\0';
  nq = ntok + 1;
  nhow_many = 0;
 
  /* now start processing */

  while( !feof(input) ) {
    /* read in a full line, allocating memory as necessary */
    textbuf[0] = '\0';
    s = textbuf;
    l = textbuf_len;
    k = 1;
    while( fgets(s, l, input) && (strlen(s) >= (l - 1)) ) {
      textbuf = realloc(textbuf, 2 * textbuf_len);
      if( !textbuf ) {
	fprintf(stderr, 
		"error: not enough memory for input line (%d bytes)\n",
		textbuf_len);
	exit(0);
      }

      s = textbuf + textbuf_len - (k++);
      l = textbuf_len;
      textbuf_len *= 2;
    }

    /* preprocesses textbuf, optionally censors it */
    if( pre_line_fun ) {
      pptextbuf = (*pre_line_fun)(textbuf);
      if( !pptextbuf ) { continue; }
    } else {
      pptextbuf = textbuf;
    }

    /* next we check to see if this line should be skipped */
    if( *pptextbuf && (!line_filter || (*line_filter)(pptextbuf)) ) {

      /* now filter some of the characters in the current line */
      if( character_filter ) { (*character_filter)(pptextbuf); }

      /* repeat for each regular expression:
	 find all the instances of a matching substring */
#if defined HAVE_LIBBOOST_REGEX

      /* damn - we compiled wide character regexes, 
	 now we've got to use them (not really, but who in their right
	 mind would compile with boost if they're not going to convert
	 to wide characters anyway) */

      if( textbuf_len > wc_textbuf_len ) {
	wc_textbuf_len = textbuf_len;
	wc_textbuf = realloc(wc_textbuf, wc_textbuf_len * sizeof(wchar_t));
	if( !wc_textbuf ) {
	  fprintf(stderr, 
		  "error: not enough memory for wide character conversion "
		  "(%d bytes)\n",
		  wc_textbuf_len * sizeof(wchar_t));
	  exit(0);
	}
      }

      /* convert as much as we can of the line into wide characters */
      s = pptextbuf;
      k = textbuf_len;
      wp = wc_textbuf;
      wclen = 0;
      /* since we ensured textbuf_len <= wctextbuf_len
	 there will never be overflow of wctextbuf below */
      while( k > 0 ) {
	l = mbrtowc(wp, s, k, &input_shiftstate);
	if( l > 0 ) {
	  wp++;
	  wclen++;
	  k -= l;
	  s += l;
	} else if( l == 0 ) {
	  break;
	} else if( l == -1 ) {
	  /* try to be robust */
	  s++; 
	  k--;
	  memset(&input_shiftstate, 0, sizeof(mbstate_t));
	} else if( l == -2) {
	  /* couldn't parse a complete character */
	  break;
	}
      }
      *wp = L'\0';

      /* now process the regex */
      for(i = 0; i < regex_count; i++) {
	k = 0;
	eflag = 0;
	/* see if a match */
	while( (k < wclen) && (regexec(&re[i].regex, wc_textbuf + k, 
				   MAX_SUBMATCH, pmatch, eflag) == 0) ) {
 	  /* all the submatches (delimited by brackets in the regex) 
 	     get converted, concatenated and the result gets word_fun'd */
	  q = tok;
	  *q++ = DIAMOND;
	  memset(&tok_shiftstate, 0, sizeof(mbstate_t));
 	  for(order = 0, z = 1; 
	      (z < MAX_SUBMATCH) && (pmatch[z].rm_so > -1); z++) {
	    if( !(re[i].submatches & (1<<z)) ) 
	      { continue; } else { order++; }
 	    /* transcribe the submatch into tok */
 	    for(j = pmatch[z].rm_so; j < pmatch[z].rm_eo; j++) {
 	      if( q < tok + MAX_TOKEN_LEN - MULTIBYTE_EPSILON ) {
		if( options & (1<<OPTION_CASEN) ) {
		  tok_len = wcrtomb(q, wc_textbuf[k + j], &tok_shiftstate); 
		} else {
		  tok_len = wcrtomb(q, towlower(wc_textbuf[k + j]), 
				    &tok_shiftstate); 
		}
		if( (tok_len > -1) ) {
		  q += tok_len;
		}
	      }
	    }
	    *q++ = DIAMOND;
	  }
 	  *q = '\0'; 

 	  /* now let each category process the token */
	  (*word_fun)(tok, order, i + 1); 

	  k += pmatch[0].rm_so + 1; /* advance string and repeat */
 	  eflag = REG_NOTBOL;
	}
      }
#else 

      /* the GNU regex routines expect an ordinary string */
      for(i = 0; i < regex_count; i++) {
	k = 0;
 	eflag = 0;
 	/* see if a match */
 	while( (k < l) && (regexec(&re[i].regex, pptextbuf + k, 
				   MAX_SUBMATCH, pmatch, eflag) == 0) ) {
 	  /* all the submatches (delimited by brackets in the regex) 
 	     get concatenated and the result gets word_fun'd */
 	  q = tok;
	  *q++ = DIAMOND;
 	  for(order = 0, z = 1; 
	      (z < MAX_SUBMATCH) && (pmatch[z].rm_so > -1); z++) {
	    if( !(re[i].submatches & (1<<z)) ) 
	      { continue; } else { order++; }
 	    /* transcribe the submatch into tok */
 	    for(j = pmatch[z].rm_so; j < pmatch[z].rm_eo; j++) {
 	      if( q < tok + MAX_TOKEN_LEN ) {
		if( options & (1<<OPTION_CASEN) ) {
		  *q++ = pptextbuf[k + j];
		} else {
		  *q++ = tolower(pptextbuf[k + j]);
		}
 	      }
 	    }
	    *q++ = DIAMOND;
 	  }
 	  *q = '\0'; 
 	  /* now let each category process the token */
	  (*word_fun)(tok, order, i + 1); 

	  k += pmatch[0].rm_so + 1; /* advance string and repeat */
 	  eflag = REG_NOTBOL;
 	}	      
      }
      
#endif

      /* default processing: reads tokens and passes them to
	 the word_fun */
      if( options & (1<<OPTION_NOREGEX) ) {
	p = pptextbuf;

	while( *p ) {
	  if( good_char(*p) ) {
	    if( nq < ntok + MAX_TOKEN_LEN ) {
	      if( !(options & (1<<OPTION_CASEN)) ) {
		*nq++ = tolower(*p);
	      } else {
		*nq++ = *p;
	      }
	    }
	  } else if( *(nq - 1) != DIAMOND ) { /* token boundary */
	    *nq++ = DIAMOND;
	    *nq = '\0';
	    if( ngram_order == 1 ) {
	      /* let each category process the token */
	      (*word_fun)(ntok, (token_order_t)1, 0);
	      nq = ntok;
	      *nq++ = DIAMOND;
	    } else {
	      if( ++nhow_many > ngram_order ) {
		nhow_many--;
		/* move all tokens down by one */
		for(nq = ntok + 1; *nq != DIAMOND; nq++) {};
		for(nq++, qq = ntok + 1; *nq; *qq++ = *nq++) {};
		*qq = '\0';
		nq = qq;
	      }

	      qq = ntok;
	      for(n = nhow_many; n > 0; n--) {
		/* let each category process the token */
		(*word_fun)(qq, (token_order_t)n, 0);
		qq++;
		/* skip to next token and repeat */
		while(*qq != DIAMOND ) { qq++; }
	      }
	    }
	  }
	  p++;
	}
      }
      
    } else { 
      /* since we don't process this line, we
	 reset XML tag state for next time */
      xml.state = TEXT; 
    }

    /* now summarize this line if required */
    if( post_line_fun ) { (*post_line_fun)(pptextbuf); }
  }
}


/***********************************************************
 * WIDE CHARACTER FILE HANDLING FUNCTIONS                  *
 * this is needed for any locale whose character set       *
 * encoding can include NUL bytes inside characters        *
 ***********************************************************/
#if defined HAVE_MBRTOWC

bool_t w_good_char(wchar_t c) {
  if( options & (1<<OPTION_CHAR_ALNUM) ) {
    return iswalnum(c);
  } else if( options & (1<<OPTION_CHAR_ALPHA) ) {
    return iswalpha(c);
  } else if( options & (1<<OPTION_CHAR_GRAPH) ) {
    return iswgraph(c);
  }
  return 0;
}

/* returns true if the line should be processed further
   depends on global mbox state */
bool_t w_mbox_line_filter(wchar_t *line) {

  bool_t line_empty = 0;
  bool_t process_line = 0; /* by default we skip the line */

  line_empty = ((line[0] == L'\n') ||
		((line[0] == L'\r') && (line[1] == L'\n'))) ? 1 : 0;

  /* below we decide if we want to process the line */
  switch(mbox.state) {
  case msUNDEF:
    if( mbox.prev_line_empty && !wcsncmp(line, L"From ", 5) ) {
      mbox.state = msHEADER;
      mbox.header.type = mbox.body.type = ctUNDEF;
      mbox.header.encoding = mbox.body.encoding = ceUNDEF;
    }
    break;
  case msHEADER:
    if( line_empty ) {
      mbox.state = msBODY;
      /* if the header was marked specially, transfer to body */
      mbox.body = mbox.header;
      /* actually, the first body after the header is always text */
      mbox.header.type = ctTEXT;
      mbox.header.encoding = ceID;
    } else {
      if( !wcsncasecmp(line, L"Content-Type:", 13) ) {
	if( mywcscasestr(line + 13, L"text") || 
	    mywcscasestr(line + 13, L"message") ||
	    mywcscasestr(line + 13, L"multipart") ) {
	  mbox.header.type = ctTEXT;
	} else {
	  mbox.header.type = ctOTHER;
	}
      } else if( !wcsncasecmp(line, L"Content-Transfer-Encoding:", 26) ) {
	if( mywcscasestr(line + 36, L"base64") ) {
	  mbox.header.encoding = ceB64;
	} else if( mywcscasestr(line + 36, L"quoted-printable") ) {
	  mbox.header.encoding = ceQP;
	} else if( mywcscasestr(line + 36, L"binary") ) {
	  mbox.header.encoding = ceBIN;
	} else {
	  mbox.header.encoding = ceID;
	}
      }
    }
    break; 
  case msBODY:
    if( mbox.prev_line_empty ) {
      if( (line[0] == L'-') && (line[1] == L'-') && !isspace(line[2]) ) {
	mbox.state = msATTACH;
	mbox.body.type = ctUNDEF;
	mbox.body.encoding = ceUNDEF;
      } else if( !wcsncmp(line, L"From ", 5) ) {
	mbox.state = msHEADER;
	mbox.header.type = mbox.body.type = ctUNDEF;
	mbox.header.encoding = mbox.body.encoding = ceUNDEF;
      }
    }
    break;
  case msATTACH:
    if( line_empty ) {
      mbox.state = msBODY;
    } else {
      if( !wcsncasecmp(line, L"Content-Type:", 13) ) {
	if( mywcscasestr(line + 13, L"text") ) {
	  mbox.body.type = ctTEXT;
	} else {
	  mbox.body.type = ctOTHER;
	}
      } else if( !wcsncasecmp(line, L"Content-Transfer-Encoding:", 26) ) {
	if( mywcscasestr(line + 36, L"base64") ) {
	  mbox.body.encoding = ceB64;
	} else if( mywcscasestr(line + 36, L"quoted-printable") ) {
	  mbox.body.encoding = ceQP;
	} else if( mywcscasestr(line + 36, L"binary") ) {
	  mbox.body.encoding = ceBIN;
	} else {
	  mbox.body.encoding = ceID;
	}
      } 
    }
    break;
  }

  mbox.prev_line_empty = line_empty; /* for next time */

  /* now decide what to do about this line */
  switch(mbox.state) {
  case msHEADER:
    if( !wcsncasecmp(line, L"From:",5) ) {
      process_line = 1;
    } else if( !wcsncasecmp(line, L"Subject:",8) ) {
      /* this is where we could remove classifier tags */
      process_line = 1;
    }
    break;
  case msBODY:
    {
      mbox.scratch.type = 
	(mbox.body.type != ctUNDEF) ? mbox.body.type : mbox.header.type;
      if( mbox.scratch.type != ctOTHER ) {
	mbox.scratch.encoding = 
	  (mbox.body.encoding != ceUNDEF) ? 
	  mbox.body.encoding : mbox.header.encoding;
	if( (mbox.scratch.encoding == ceID) || 
	    (mbox.scratch.encoding == ceQP) ) {
	  process_line = 1;
	}
      }
    }
    break;
  case msATTACH:
    /* MIME section headers are always processed, since they describe content */
    process_line = 1;
    break;
  default:
    /* nothing */
    break;
  }

  return process_line;
}

/* removes tags in the string - modifies in place */
/* the name of this function is a misnomer, since it doesn't
   parse xml properly. But we just want a simple kludge 
   for most html flavours */
void w_xml_character_filter(wchar_t *line) {

  wchar_t *q;

  q = line;
  while( *line ) {
    switch(xml.state) {
    case TEXT:
      /* does it look like <x where x is either alpha or punctuation? */
      if( line[0] == L'<' ) {
	if( !wcsncmp(line + 1, L"!--", 3) ) {
	  xml.state = COMMENT;
	  line += 3;
	} else if( wcsncasecmp(line + 1, L"script", 6) != 0 ) {
	  xml.state = SPECIAL;
	  line += 6;
	} else if( !wcsncasecmp(line + 1, L"br>", 3) ) {
	  *q++ = ' ';
	  line += 3;
	} else if( iswalpha(line[1]) || iswpunct(line[1]) ) {
	  xml.state = TAG;
	  line += 1;
	} else {
	  *q++ = *line; 
	}
      } else {
	*q++ = *line;
      }
      break;
    case TAG:
      if( line[1] == L'>' ) {
	xml.state = TEXT;
      }
      break;
    case COMMENT:
      if( line[0] == L'-' && !wcsncmp(line + 1, L"->", 2) ) {
	xml.state = TEXT;
	line += 2;
      }
      break;
    case SPECIAL:
      if( line[0] == L'<' ) {
	if( !wcsncasecmp(line + 1, L"/script", 7) ) {
	  xml.state = TEXT;
	  line += 7;
	}
      }
      break;
    }
    line++;
  }
  *q = L'\0'; /* mark the end of the clean text string */
}

/* reads a text file as input, converting each line
into a wide character representation and applies several
filters. */
void w_process_file(FILE *input, 
		    int (*line_filter)(wchar_t *),
		    void (*character_filter)(wchar_t *), 
		    void (*word_fun)(char *, token_order_t, regex_count_t), 
		    char *(*pre_line_fun)(char *),
		    void (*post_line_fun)(char *)) {
  char *s, *pptextbuf;
  regex_count_t i;
  charbuf_len_t k;
  int eflag;
  token_order_t z, j, n, order;
  charbuf_len_t l;
  charbuf_len_t wclen;

  wchar_t *wp;
  mbstate_t tok_shiftstate, input_shiftstate;
  charbuf_len_t tok_len;
  char *q, *qq;
  char tok[MAX_TOKEN_LEN+2];

  regmatch_t pmatch[MAX_SUBMATCH];

  char *nq;
  char ntok[MAX_TOKEN_LEN+2];
  token_order_t nhow_many;

  /* initialize the norex state */
  ntok[0] = DIAMOND;
  ntok[1] = '\0';
  nq = ntok + 1;
  nhow_many = 0;

  memset(&input_shiftstate, 0, sizeof(mbstate_t));

  while( !feof(input) ) {

    /* read in a full line, allocating memory as necessary */
    textbuf[0] = '\0';
    s = textbuf;
    l = textbuf_len;
    k = 1;
    while( fgets(s, l, input) && (strlen(s) >= (l - 1)) ) {
      textbuf = realloc(textbuf, 2 * textbuf_len);
      if( !textbuf ) {
	fprintf(stderr, 
		"error: not enough memory for input line (%d bytes)\n",
		textbuf_len);
	exit(0);
      }

      s = textbuf + textbuf_len - (k++);
      l = textbuf_len;
      textbuf_len *= 2;
    }

    /* preprocesses textbuf, optionally censors it */
    if( pre_line_fun ) {
      pptextbuf = (*pre_line_fun)(textbuf);
      if( !pptextbuf ) { continue; }
    } else {
      pptextbuf = textbuf;
    }

    /* now convert the line into a wide character string */
    if( textbuf_len > wc_textbuf_len ) {
      wc_textbuf_len = textbuf_len;
      wc_textbuf = realloc(wc_textbuf, wc_textbuf_len * sizeof(wchar_t));
      if( !wc_textbuf ) {
	fprintf(stderr, 
		"error: not enough memory for wide character conversion "
		"(%ld bytes)\n",
		(long int)(wc_textbuf_len * sizeof(wchar_t)));
	exit(0);
      }
    }

    /* convert as much as we can of the line into wide characters */
    s = pptextbuf;
    k = textbuf_len;
    wp = wc_textbuf;
    wclen = 0;
    /* since we ensured textbuf_len <= wctextbuf_len
       there will never be overflow of wctextbuf below */
    while( k > 0 ) {
      l = mbrtowc(wp, s, k, &input_shiftstate);
      if( l > 0 ) {
	wp++;
	wclen++;
	k -= l;
	s += l;
      } else if( l == 0 ) {
	break;
      } else if( l == -1 ) {
	/* try to be robust */
	s++; 
	k--;
	memset(&input_shiftstate, 0, sizeof(mbstate_t));
      } else if( l == -2) {
	/* couldn't parse a complete character */
	break;
      }
    }
    *wp = L'\0';

    /* next we check to see if this line should be skipped */
    if( (wclen > 0) && 
	(!line_filter || (*line_filter)(wc_textbuf)) ) {
      /* now filter some of the characters in the current line */
      if( character_filter ) { (*character_filter)(wc_textbuf); }

      /* repeat for each regular expression:
	 find all the instances of a matching substring */
#if defined HAVE_LIBBOOST_REGEX

      for(i = 0; i < regex_count; i++) {
	k = 0;
	eflag = 0;
	/* see if a match */
	while( (k < wclen) && (regexec(&re[i].regex, wc_textbuf + k, 
				   MAX_SUBMATCH, pmatch, eflag) == 0) ) {
 	  /* all the submatches (delimited by brackets in the regex) 
 	     get converted, concatenated and the result gets word_fun'd */
	  q = tok;
	  *q++ = DIAMOND;
	  memset(&tok_shiftstate, 0, sizeof(mbstate_t));
 	  for(order = 0, z = 1; 
	      (z < MAX_SUBMATCH) && (pmatch[z].rm_so > -1); z++) {
	    if( !(re[i].submatches & (1<<z)) ) 
	      { continue; } else { order++; }
 	    /* transcribe the submatch into tok */
 	    for(j = pmatch[z].rm_so; j < pmatch[z].rm_eo; j++) {
 	      if( q < tok + MAX_TOKEN_LEN - MULTIBYTE_EPSILON ) {
		if( options & (1<<OPTION_CASEN) ) {
		  tok_len = wcrtomb(q, wc_textbuf[k + j], &tok_shiftstate); 
		} else {
		  tok_len = wcrtomb(q, towlower(wc_textbuf[k + j]), 
				    &tok_shiftstate); 
		}
		if( (tok_len > -1) ) {
		  q += tok_len;
		}
	      }
	    }
	    *q++ = DIAMOND;
	  }
 	  *q = '\0'; 

 	  /* now let each category process the token */
 	  (*word_fun)(tok, order, i + 1); 

	  k += pmatch[0].rm_so + 1; /* advance string and repeat */
 	  eflag = REG_NOTBOL;
	}
      }
#else 
      /* the GNU regex routines expect a multibyte string */
      if( textbuf_len > aux_textbuf_len ) {
	aux_textbuf_len = textbuf_len;
	aux_textbuf = realloc(aux_textbuf, aux_textbuf_len);
	if( !aux_textbuf ) {
	  fprintf(stderr, 
		  "error: not enough memory for auxiliary text buffer "
		  "(%d bytes)\n",
		  aux_textbuf_len);
	  exit(0);
	}
      }
      l = wcstombs(aux_textbuf, wc_textbuf, aux_textbuf_len - 1);
      aux_textbuf[aux_textbuf_len - 1] = '\0';

      for(i = 0; i < regex_count; i++) {
	k = 0;
 	eflag = 0;
 	/* see if a match */
 	while( (k < l) && (regexec(&re[i].regex, aux_textbuf + k, 
				   MAX_SUBMATCH, pmatch, eflag) == 0) ) {
 	  /* all the submatches (delimited by brackets in the regex) 
 	     get concatenated and the result gets word_fun'd */
 	  q = tok;
	  *q++ = DIAMOND;
 	  for(order = 0, z = 1; 
	      (z < MAX_SUBMATCH) && (pmatch[z].rm_so > -1); z++) {
	    if( !(re[i].submatches & (1<<z)) ) 
	      { continue; } else { order++; }
 	    /* transcribe the submatch into tok */
 	    for(j = pmatch[z].rm_so; j < pmatch[z].rm_eo; j++) {
 	      if( q < tok + MAX_TOKEN_LEN ) {
		if( options & (1<<OPTION_CASEN) ) {
		  *q++ = aux_textbuf[k + j];
		} else {
		  *q++ = tolower(aux_textbuf[k + j]);
		}
 	      }
 	    }
	    *q++ = DIAMOND; /* hope DIAMOND can never be matched inside regex */
 	  }
 	  *q = '\0'; 
 	  /* now let each category process the token */
	  (*word_fun)(tok, order, i + 1); 

	  k += pmatch[0].rm_so + 1; /* advance string and repeat */
 	  eflag = REG_NOTBOL;
 	}	      
      }
      
#endif

      /* default processing: reads tokens, converting them
	 to multibyte representation before passing them to
	 the word_fun */
      if( options & (1<<OPTION_NOREGEX) ) {
	wp = wc_textbuf;
	memset(&tok_shiftstate, 0, sizeof(mbstate_t));

	while( *wp ) {
	  if( w_good_char(*wp) ) {
	    if( !(options & (1<<OPTION_CASEN)) ) {
	      *wp = towlower(*wp);
	    }

	    if( nq < ntok + MAX_TOKEN_LEN - MULTIBYTE_EPSILON ) {
	      tok_len = wcrtomb(nq, *wp, &tok_shiftstate); 
	      if( (tok_len > -1) ) {
		nq += tok_len;
	      }
	    }
	  } else if( *(nq - 1) != DIAMOND ) { /* token boundary */
	    *nq++ = DIAMOND;
	    *nq = '\0';
	    if( ngram_order == 1 ) {
	      /* let each category process the token */
	      (*word_fun)(ntok, (token_order_t)1, 0);
	      nq = ntok;
	      *nq++ = DIAMOND;
	    } else {
	      if( ++nhow_many > ngram_order ) {
		nhow_many--;
		/* move all tokens down by one */
		for(nq = ntok + 1; *nq != DIAMOND; nq++) {};
		for(nq++, qq = ntok + 1; *nq; *qq++ = *nq++) {};
		*qq = '\0';
		nq = qq;
	      }

	      qq = ntok;
	      for(n = nhow_many; n > 0; n--) {
		/* let each category process the token */
		(*word_fun)(qq, (token_order_t)n, 0);
		qq++;
		/* skip to next token and repeat */
		while(*qq != DIAMOND ) { qq++; }
	      }
	    }
	  }
	  wp++;
	}
      }

    } else { 
      /* since we don't process this line, we
	 reset XML tag state for next time */
      xml.state = TEXT; 
    }

    /* now summarize this line if required */
    if( post_line_fun ) { (*post_line_fun)(pptextbuf); }
      
  }
}

#endif /* DISABLE_WCHAR */
